# $Id: smb.pac 3239 2006-06-08 02:46:53Z vern $
#
# CIFS/SMB

# TODO:
# - Built support for unicode strings
# - Unicode as an implicit attribute (as byteorder)
# - &truncation_ok attribute for the last field of a record to deal with partial data

%include bro.pac

analyzer SMB withcontext {};

# let byteorder = littleendian;

type SMB_dos_error = record {
	error_class	: uint8;
	reserved	: uint8;
	error		: uint16;
};

type SMB_error (err_status_type: int) = case err_status_type of {
	0 -> dos_error: SMB_dos_error;
	1 -> status: int32;
};

type SMB_header = record {
	protocol	: bytestring &length = 4;
	command		: uint8;
	status		: SMB_error(err_status_type);
	flags		: uint8;
	flags2		: uint16;
	pad		: padding[12];
	tid		: uint16;
	pid		: uint16;
	uid		: uint16;
	mid		: uint16;
} &let {
	err_status_type = (flags2 >> 14) & 1;
	unicode = (flags2 >> 15) & 1;
} &byteorder = littleendian;

# TODO: compute this as
# let smb_header_length = sizeof(SMB_header);
let smb_header_length = 32;

type SMB_body = record {
	word_count	: uint8;
	parameter_words : uint16[word_count];
	byte_count	: uint16;
	# buffer	: uint8[byte_count];
} &let {
	body_length = 1 + word_count * 2 + 2 + byte_count;
} &byteorder = littleendian;

type SMB_ascii_string		= uint8[] &until($element == 0);
type SMB_unicode_string(offset: int) = record {
	pad	: padding[offset & 1];
	s	: uint16[] &until($element == 0);
};

type SMB_string(unicode: bool, offset: int) = case unicode of {
	true	-> u: SMB_unicode_string(offset);
	false	-> a: SMB_ascii_string;
};

type SMB_andx = record {
	command		: uint8;
	reserved	: uint8;
	offset		: uint16;
};

type SMB_generic_andx = record {
	word_count	: uint8;
	andx_u		: case word_count of {
		0	-> null : empty;
		default -> andx	: SMB_andx;
	};
} &byteorder = littleendian;

type SMB_dialect = record {
	bufferformat	: uint8;	# must be 0x2
	dialectname	: SMB_ascii_string;
};

type SMB_negotiate = record {
	word_count	: uint8;	# must be 0
	byte_count	: uint16;
	dialects	: SMB_dialect[] &length = byte_count;
} &byteorder = littleendian;

type SMB_tree_connect_andx(unicode: bool) = record {
	word_count	: uint8;
	andx		: SMB_andx;
	flags		: uint16;
	password_length	: uint16;
	byte_count	: uint16;
	password	: uint8[password_length];
	path		: SMB_string(unicode, offsetof(path));
	service		: SMB_ascii_string;
} &byteorder = littleendian;

type SMB_nt_create_andx(unicode: bool) = record {
	word_count	: uint8;
	andx		: SMB_andx;
	reserved	: uint8;
	name_length	: uint16;
	flags		: uint32;
	rest_words	: uint8[word_count * 2 - 11];
	byte_count	: uint16;
	name		: SMB_string(unicode, offsetof(name));
} &byteorder = littleendian;

type SMB_read_andx = record {
	word_count	: uint8;
	andx		: SMB_andx;
	fid		: uint16;
	offset		: uint32;
	max_count	: uint16;
	min_count	: uint16;
	max_count_high	: uint16;
	remaining	: uint16;
	offset_high	: uint32;
	byte_count	: uint16;
} &byteorder = littleendian;

type SMB_read_andx_response = record {
	word_count	: uint8;
	andx		: SMB_andx;
	remaining	: uint16;
	data_compact	: uint16;
	reserved	: uint16;
	data_len	: uint16;
	data_offset	: uint16;
	data_len_high	: uint16;
	reserved2	: uint16[4];
	byte_count	: uint16;
	pad		: padding to data_offset - smb_header_length;
	data		: bytestring &length = data_length;
} &let {
	data_length = data_len_high * 0x10000 + data_len;
} &byteorder = littleendian;

type SMB_write_andx = record {
	word_count	: uint8;
	andx		: SMB_andx;
	fid		: uint16;
	offset		: uint32;
	reserved	: uint32;
	write_mode	: uint16;
	remaining	: uint16;
	data_len_high	: uint16;
	data_len	: uint16;
	data_offset	: uint16;
	rest_words	: uint8[word_count * 2 - offsetof(rest_words) + 1];
	byte_count	: uint16;
	pad		: padding to data_offset - smb_header_length;
	data		: bytestring &length = data_length;
} &let {
	data_length = data_len_high * 0x10000 + data_len;
} &byteorder = littleendian;

type SMB_write_andx_response = record {
	word_count	: uint8;
	andx		: SMB_andx;
	count		: uint16;	# written bytes
	remaining	: uint16;
	reserved	: uint32;
	byte_count	: uint16;
} &byteorder = littleendian;

type SMB_transaction(trans_type: int, unicode: bool) = record {
	word_count		: uint8;
	total_param_count	: uint16;
	total_data_count	: uint16;
	max_param_count		: uint16;
	max_data_count		: uint16;
	max_setup_count		: uint8;
	reserved		: uint8;
	flags			: uint16;
	timeout			: uint32;
	reserved2		: uint16;
	param_count		: uint16;
	param_offset		: uint16;
	data_count		: uint16;
	data_offset		: uint16;
	setup_count		: uint8;
	reserved3		: uint8;
	setup			: uint16[setup_count];
	byte_count		: uint16;
	name_u			: case trans_type of {
		1 -> name: SMB_string(unicode, offsetof(name_u));
		2 -> null: empty;
		};
	pad0			: padding to param_offset - smb_header_length;
	parameters		: bytestring &length = param_count;
	pad1			: padding to data_offset - smb_header_length;
	data			: bytestring &length = data_count;
} &byteorder = littleendian;

type SMB_transaction_secondary(unicode: bool) = record {
	word_count		: uint8;
	total_param_count	: uint16;
	total_data_count	: uint16;
	param_count		: uint16;
	param_offset		: uint16;
	param_displacement	: uint16;
	data_count		: uint16;
	data_offset		: uint16;
	data_displacement	: uint16;
	fid			: uint16;
	byte_count		: uint16;
	pad0			: padding to param_offset - smb_header_length;
	parameters		: bytestring &length = param_count;
	pad1			: padding to data_offset - smb_header_length;
	data			: bytestring &length = data_count;
} &byteorder = littleendian;

type SMB_transaction_response(unicode: bool) = record {
	word_count		: uint8;
	total_param_count	: uint16;
	total_data_count	: uint16;
	reserved		: uint16;
	param_count		: uint16;
	param_offset		: uint16;
	param_displacement	: uint16;
	data_count		: uint16;
	data_offset		: uint16;
	data_displacement	: uint16;
	setup_count		: uint8;
	reserved2		: uint8;
	setup			: uint16[setup_count];
	byte_count		: uint16;
	pad0			: padding to param_offset - smb_header_length;
	parameters		: bytestring &length = param_count;
	pad1			: padding to data_offset - smb_header_length;
	data			: bytestring &length = data_count;
} &byteorder = littleendian;

type SMB_get_dfs_referral(unicode: bool) = record {
	max_referral_level	: uint16;
	file_name		: SMB_string(unicode, offsetof(file_name));
} &byteorder = littleendian;
